/*
    wavedata.h - Structures to represent a set of wavetables

    Copyright (C) 2002  Mike Rawes

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef MYK_WAVEDATA_H
#define MYK_WAVEDATA_H

#include "ladspa.h"

typedef enum
{
    YES,
    NO,
    OUT
} XFade;

/*
 * Structure holding a single segment of sample data
 * along with information relating to playback.
 */
typedef struct
{
    XFade fade;                         /* Cross fading or otherwise */
    unsigned long count;                /* Sample count */
    LADSPA_Data phase_scale_factor;     /* Phase scale factor for playback */
    LADSPA_Data *samples_base;          /* Sample data played back over whole
                                            pitch range */
    LADSPA_Data *samples_extra;         /* Sample data played back at decreasing
                                            amplitude as pitch increases */
    LADSPA_Data min_frequency;          /* Minimum pitch */
    LADSPA_Data max_frequency;          /* Maximum pitch */
    LADSPA_Data range_scale_factor;     /* Blend scale factor for cross fading */
} Wavetable;

/*
 * Structure holding the wavetable data
 */
typedef struct
{
    unsigned long number_of_wavetables;     /* Number of wavetables in wavedata */
    Wavetable **wavetables;                 /* One or more wavetables */
    Wavetable *wtbl_lf;                     /* Wavetable for frequencies lower than
                                                lowest wavetable min frequency */
    Wavetable *wtbl_hf;                     /* Wavetable for frequencies between
                                                highest wavetable max frequency
                                                and 0.99 * Nyquist */
    Wavetable *wtbl_fadeout;                /* Wavetable for frequencies between
                                                0.99 * Nyquist and Nyquist - this
                                                fades out to zero at Nyquist */
    unsigned long lookup_min;               /* Smallest key in lookup table */
    unsigned long lookup_max;               /* Largest key in lookup table */
    unsigned long *lookup;                  /* Wavetable lookup vector */
    LADSPA_Data sample_rate;                /* Sample rate */
} Wavedata;

/*
 * Structure for representing contents of wavedata header file
 */
typedef struct
{
    unsigned long number_of_wavetables;
    unsigned long lowest_harmonic;
    unsigned long highest_harmonic;
    unsigned long *max_harmonics;
    unsigned long *sample_counts;
    LADSPA_Data *sample_data_base;
    LADSPA_Data *sample_data_extra;
} Wdatheader;

#ifdef __cplusplus
extern "C" {
#endif

Wavedata *initWavedata(Wdatheader *wavedata, unsigned long sample_rate);
void destroyWavedata(Wavedata *wdat);

#ifdef __cplusplus
} /* extern "C" { */
#endif

#endif /* MYK_WAVEDATA_H */
