#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "common.h"
#include "wavedata.h"
#include "ladspa.h"

/*
 * Intialise wavedata from header
 */
Wavedata *
initWavedata(Wdatheader *wavedata, unsigned long sample_rate)
{
    unsigned long wavetable_count;
    unsigned long lowest_harmonic;
    unsigned long highest_harmonic;
    unsigned long *max_harmonics;
    unsigned long *sample_counts;
    LADSPA_Data *samples_base;
    LADSPA_Data *samples_extra;

    Wavedata *wdat;
    Wavetable *wtbl;
    unsigned long max_harmonic;
    LADSPA_Data min_frequency, max_frequency;
    unsigned long w;
    unsigned long s, count;
    unsigned long lookup_size;
    unsigned long i, i_max;
    size_t bytesize;

/* Get information from wavedata header */
/* Argh. This is sooo *ugly*! */
    wavetable_count = wavedata->number_of_wavetables;
    lowest_harmonic = wavedata->lowest_harmonic;
    highest_harmonic = wavedata->highest_harmonic;
    max_harmonics = wavedata->max_harmonics;
    sample_counts = wavedata->sample_counts;
    samples_base = wavedata->sample_data_base;
    samples_extra = wavedata->sample_data_extra;

/* Attempt to create new wavedata */
    wdat = (Wavedata *)malloc(sizeof(Wavedata));
    if (!wdat) {
        return NULL;
    }

/* Allocate wavetables */
    bytesize = sizeof(Wavetable *) * ((size_t)wavetable_count);
    wdat->wavetables = (Wavetable **)malloc(bytesize);
    if (!wdat->wavetables) {
        free(wdat);
        return NULL;
    }

    wdat->sample_rate = (LADSPA_Data)sample_rate;
    wdat->number_of_wavetables = wavetable_count;
    wdat->lookup_min = max_harmonics[wavetable_count - 1];
    wdat->lookup_max = highest_harmonic - 1;

/* Initialise wavetable lookup vector */
    lookup_size = wdat->lookup_max - wdat->lookup_min + 1;
    bytesize = (size_t)lookup_size * sizeof(unsigned long);
    wdat->lookup = (unsigned long *)malloc(bytesize);
    if (!wdat->lookup) {
        free(wdat->wavetables);
        free(wdat);
        return NULL;
    }

/* Set up wavetables */
    min_frequency = wdat->sample_rate / 2.0f / (LADSPA_Data)highest_harmonic;

    for (w = 0; w < wdat->number_of_wavetables; w++) {
        max_harmonic = max_harmonics[w];
        max_frequency = wdat->sample_rate / 2.0f / (LADSPA_Data)max_harmonic;

        wdat->wavetables[w] = (Wavetable *)malloc(sizeof(Wavetable));
        if (!wdat->wavetables[w]) {
        /* Oops */
            destroyWavedata(wdat);
            return NULL;
        }

        count = sample_counts[w];

        wtbl = wdat->wavetables[w];
        wtbl->fade = YES;
        wtbl->count = count - 3; /* Don't count the duplicates */
        wtbl->phase_scale_factor = (LADSPA_Data)(wtbl->count) / wdat->sample_rate;
        wtbl->min_frequency = min_frequency;
        wtbl->max_frequency = max_frequency;
        wtbl->range_scale_factor = 1.0f / (max_frequency - min_frequency);
        wtbl->samples_base = samples_base;
        wtbl->samples_extra = samples_extra;
        samples_base += count;
        samples_extra += count;

        min_frequency = max_frequency;
    }

/* Extra wavetables played back at frequency extremes */
/*
 * For frequencies below first wavetable's minimum frequency
 * No crossfading is done
 */
    w = 0;
    wtbl = (Wavetable *)malloc(sizeof(Wavetable));
    if (!wtbl) {
    /* Bugger */
        destroyWavedata(wdat);
        return NULL;
    }
    wtbl->fade = NO;
    wtbl->count = wdat->wavetables[w]->count;
    wtbl->phase_scale_factor = (LADSPA_Data)(wtbl->count) / wdat->sample_rate;
/* The next three aren't used, but are here for completeness */
    wtbl->min_frequency = 0.0f;
    wtbl->max_frequency = wdat->wavetables[w]->min_frequency;
    wtbl->range_scale_factor = 1.0f / (wtbl->max_frequency - wtbl->min_frequency);
/*
 * Both base and extra samples are played back at max amplitude
 * so combine base and extra samples from first wavetable, since
 * we don't need to do any fading
 */
/* Get real count - we don't ignore extra samples here */
    count = sample_counts[w];
    bytesize = (size_t)count * sizeof(LADSPA_Data);
    wtbl->samples_base = (LADSPA_Data *)malloc(bytesize);
    if (!wtbl->samples_base) {
    /* Dang. */
        destroyWavedata(wdat);
        return NULL;
    }
    samples_base = wdat->wavetables[w]->samples_base;
    samples_extra = wdat->wavetables[w]->samples_extra;
    for (s = 0; s < count; s++) {
        wtbl->samples_base[s] = samples_base[s] + samples_extra[s];
    }
    wtbl->samples_extra = NULL;
    wdat->wtbl_lf = wtbl;

/*
 * For frequencies above last wavetable's maximum frequency,
 * and below 0.99 * Nyquist
 * No crossfading is done
 */
    w = wdat->number_of_wavetables - 1;
    wtbl = (Wavetable *)malloc(sizeof(Wavetable));
    if (!wtbl) {
        destroyWavedata(wdat);
        return NULL;
    }
    wtbl->fade = NO;
    wtbl->count = wdat->wavetables[w]->count;
    wtbl->phase_scale_factor = (LADSPA_Data)(wtbl->count) / wdat->sample_rate;
/* The next three aren't used, but are here for completeness */
    wtbl->min_frequency = wdat->wavetables[w]->max_frequency;
    wtbl->max_frequency = 0.99f * wdat->sample_rate / 2.0f;
    wtbl->range_scale_factor = 1.0f / (wtbl->max_frequency - wtbl->min_frequency);
/* Base samples are the same as those of last table */
    wtbl->samples_base = wdat->wavetables[w]->samples_base;
    wtbl->samples_extra = NULL;
    wdat->wtbl_hf = wtbl;

/* For frequencies above 0.99 * Nyquist and below Nyquist */
    wtbl = (Wavetable *)malloc(sizeof(Wavetable));
    if (!wtbl) {
        destroyWavedata(wdat);
        return NULL;
    }
    wtbl->fade = OUT;
    wtbl->count = wdat->wavetables[w]->count;
    wtbl->phase_scale_factor = (LADSPA_Data)(wtbl->count) / wdat->sample_rate;
/* These three are used for fading out */
    wtbl->min_frequency = 0.99f * wdat->sample_rate / 2.0f;
    wtbl->max_frequency = wdat->sample_rate / 2.0f;
    wtbl->range_scale_factor = 1.0f / (wtbl->max_frequency - wtbl->min_frequency);
/* Only one set of samples for playback - same as wtbl_hf */
    wtbl->samples_base = wdat->wtbl_hf->samples_base;
    wtbl->samples_extra = NULL;
    wdat->wtbl_fadeout = wtbl;

/* Add lookup data to vector */
    i_max = highest_harmonic;
    for (w = 0; w < wdat->number_of_wavetables; w++) {
        max_harmonic = max_harmonics[w];
        for (i = max_harmonic; i < i_max; i++) {
            wdat->lookup[i - wdat->lookup_min] = w;
        }
        i_max = max_harmonic;
    }

    return wdat;
}

/*
 * Free all allocated data
 * Checking is done before free, as this is called
 * by initWavedata if it fails at any point
 */
void
destroyWavedata(Wavedata *wdat)
{
    Wavetable *wtbl;
    unsigned long w;

    if (wdat) {
        for (w = 0; w < wdat->number_of_wavetables; w++) {
            wtbl = wdat->wavetables[w];
            if (wtbl) free(wtbl);
        }
    /* This table may have base samples allocated */
        wtbl = wdat->wtbl_lf;
        if (wtbl) {
            if (wtbl->samples_base) free(wtbl->samples_base);
            free(wtbl);
        }
    /* These tables' base samples just point to those of the last wavetable */
        if (wdat->wtbl_hf) free(wdat->wtbl_hf);
        if (wdat->wtbl_fadeout) free(wdat->wtbl_fadeout);
        free(wdat->lookup);
        free(wdat);
    }
}
